(* $Id: godi_file.ml 654 2006-10-08 21:25:02Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Unix


let uquote = Filename.quote;;


type split_filename =
  | Absolute of string list
  | Relative of string list

let pathsep_re = Str.regexp "/" 

let split_filename s =
  let l = Str.split_delim pathsep_re s in
  match l with
    | "" :: path ->
	Absolute (List.filter (fun s -> s <> "") path)
    | path ->
	Relative (List.filter (fun s -> s <> "") path)

let combine_filename =
  function
    | Absolute path ->
	"/" ^ String.concat "/" path
    | Relative path ->
	String.concat "/" path


exception Not_a_prefix

let strip_prefix ~prefix file =
  let rec strip p f =
    match (p, f) with
      | (cp :: p'), (cf :: f') when cp = cf ->
	  strip p' f'
      | [], _ ->
	  f
      | _ ->
	  raise Not_a_prefix
  in

  match (prefix,file) with
    | Absolute p, Absolute f ->
	strip p f
    | Relative p, Relative f ->
	strip p f
    | _ ->
	raise Not_a_prefix


let root = "/"


let catch_unix_error f arg =
  try f arg
  with
      Unix_error(err,_,errarg) ->
	let errarg' = if errarg <> "" then ": " ^ errarg else errarg in
	raise(Sys_error(Unix.error_message err ^ errarg'))
;;


type file_id =
    { file_device : int;
      file_inode : int;
    }

let file_id file =
  let s = stat file in
  { file_inode = s.st_ino;
    file_device = s.st_dev;
  }
;;


let same_file f1 f2 =
  (* Returns [true] if both files exist, and both files are the same *)
  try
    let s1 = stat f1 in
    let s2 = stat f2 in
    s1.st_dev = s2.st_dev && s1.st_ino = s2.st_ino
  with
      Unix_error(_,_,_) -> false
;;


type file_fingerprint =
    { file_mtime : float;
      file_ctime : float;
      file_size : int;
    }
;;

let file_fingerprint f =
  let s = stat f in
  { file_mtime = s.st_mtime;
    file_ctime = s.st_ctime;
    file_size = s.st_size;
  }
;;


let read_dir ?(include_dots=false) =
  (* Sys.readdir does not return ".", and "..", so we do not use it *)
  catch_unix_error
    (fun d ->
       let dh = Unix.opendir d in
       let f = ref [] in
       try
	 while true do
	   let n = Unix.readdir dh in
	   if (n <> "." && n <> "..") || include_dots then
	     f := n :: !f;
	 done;
	 assert false
       with
	   End_of_file ->
	     Unix.closedir dh;
	     List.rev !f
	 | error ->
	     Unix.closedir dh;   (* Cannot read entry *)
	     raise error
    )
;;


let file_type ?(is_regular = false) 
              ?(is_dir = false) 
	      ?(is_existing = false)
	      ?(is_symlink = false) 
	      ?(is_not_symlink = false) 
	      fn =
  let lkind =
    try
      Some ((Unix.lstat fn).Unix.st_kind) 
    with
	Unix.Unix_error(_,_,_) -> None in
  let kind =
    try
      Some ((Unix.stat fn).Unix.st_kind) 
    with
	Unix.Unix_error(_,_,_) -> None in
  (* Negative logic: *)
  match lkind, kind with
      Some Unix.S_REG, _ -> not is_dir && not is_symlink
    | Some Unix.S_DIR, _ -> not is_regular && not is_symlink
    | Some Unix.S_LNK, Some Unix.S_REG -> not is_dir && not is_not_symlink
    | Some Unix.S_LNK, Some Unix.S_DIR -> not is_regular && not is_not_symlink
    | Some Unix.S_LNK, Some _ -> 
	not is_regular && not is_dir && not is_not_symlink
    | Some Unix.S_LNK, None ->
	not is_regular && not is_dir && not is_existing && not is_not_symlink
    | Some _, _ ->
	not is_regular && not is_dir && not is_symlink
    | None, _ ->
	not is_regular && not is_dir && not is_existing && not is_symlink
;;


let file_exists ?(symlink=false) fn =
  file_type ~is_existing:true fn || (symlink && file_type ~is_symlink:true fn)
;;


let file_is_dir fn =
  file_type ~is_dir:true fn ;;


	    
let file_mtime =
  catch_unix_error
    (fun fn ->
       (Unix.stat fn).Unix.st_mtime)
;;


	    
let file_size =
  catch_unix_error
    (fun fn ->
       (Unix.LargeFile.stat fn).Unix.LargeFile.st_size)
;;


let read_dir_hierarchy ?(include_dirs=false) root =
  let rec collect rel_path =
    let path = Filename.concat root rel_path in
    let files = read_dir path in
    List.flatten
      (List.map
	 (fun file ->
	    let p = Filename.concat path file in
	    let r = Filename.concat rel_path file in
	    if file_type ~is_dir:true ~is_not_symlink:true p then (
	      let sub_dir = collect r in
	      if include_dirs then r :: sub_dir else sub_dir
	    )
	    else [ r ]
	 )
	 files)
  in
  collect ""
;;


let read_file_bin filename =
  (* Return the contents of [filename] *)
  let f = open_in_bin filename in
  try
    let n = in_channel_length f in
    let s = String.create n in
    really_input f s 0 n;
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let crlf_re = Str.regexp "\r\n";;
let lf_re = Str.regexp "\n";;

let read_file ~mode filename =
  let s = read_file_bin filename in
  match mode with
    | `Binary -> s
    | `Text -> Str.global_replace crlf_re "\n" s
;;


let read_file_lines filename =
  let rec fixup_eof lines =
    match lines with
      | [ "" ] -> []   (* Last line ends with LF *)
      | [] -> []
      | line :: lines -> line :: fixup_eof lines
  in
  fixup_eof (Str.split_delim lf_re (read_file ~mode:`Text filename))
;;


let write_file ~mode filename contents =
  let f = open_out_gen [ Open_wronly; Open_creat; Open_trunc;
			 ( match mode with
			     | `Text -> Open_text
			     | `Binary -> Open_binary
			 )
		       ] 0o666 filename in
  try
    output_string f contents;
    close_out f
  with
    | error -> close_out f; raise error
;;


let write_file_lines filename lines =
  let f = 
    open_out_gen 
      [ Open_wronly; Open_creat; Open_trunc; Open_text ] 0o666 filename in
  try
    List.iter
      (fun line ->
	 output_string f line;
	 output_string f "\n"
      )
      lines;
    close_out f
  with
    | error -> close_out f; raise error
;;


let rec remove ?(recurse = false) =
  catch_unix_error
    (fun d ->
       if recurse then (
	 if file_type ~is_dir:true ~is_not_symlink:true d then (
	   let l = read_dir d in
	   List.iter
	     (fun f ->
		let df = Filename.concat d f in
		if f <> "." && f <> ".." then (
		  remove ~recurse df
		)
	     )
	     l;
	   Unix.rmdir d
	 )
	 else
	   if file_exists ~symlink:true d then
	     Sys.remove d
       )
       else
	 if file_type ~is_dir:true ~is_not_symlink:true d then
	   Unix.rmdir d
	 else
	   Sys.remove d
    )
;;


let remove_dir ?(recurse = false) fn =
  if file_type ~is_dir:true ~is_not_symlink:true fn then
    remove ~recurse fn
  else
    raise(Sys_error("Not a directory: " ^ fn))
;;


let rec create_dir ?(recurse = false) =
  catch_unix_error
    (fun d ->
       if recurse then (
	 try
	   Unix.mkdir d 0o777
	 with
	     Unix.Unix_error(Unix.EEXIST,_,_) ->
	       (* Only check that it is already a directory: *)
	       if not (file_type ~is_dir:true d) then
		 raise(Sys_error ("Not a directory: " ^ d))
	   | Unix.Unix_error(Unix.ENOENT,_,_) ->
	       (* Try to ensure that the parent exists: *)
	       let d' = Filename.dirname d in
	       create_dir ~recurse:true d';
	       Unix.mkdir d 0o777
       ) else
	 Unix.mkdir d 0o777
    )
;;


exception Not_a_symlink

let read_symlink file =
  try
    Unix.readlink file
  with
    | Unix.Unix_error(Unix.EINVAL,_,_) -> raise Not_a_symlink
    | e -> catch_unix_error (fun () -> raise e) ()


let create_symlink src dest =
  catch_unix_error (Unix.symlink src) dest


let tmpdir_dir = ref None;;
let tmpdir_count = ref 0 ;;

type tmpdir = string ;;

let tmpdir_path t = t ;;

let rec create_tdir prefix n =
  try
    let name = prefix ^ string_of_int n in
    Unix.mkdir name 0o777;
    name
  with
      Unix.Unix_error(Unix.EEXIST,_,_) ->
	create_tdir prefix (n+1)
;;


let create_tmp_dir() =
  let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
  catch_unix_error (create_tdir tmp_prefix) 0
;;


let allocate_tmpdir() =
  (* Ensure tmpdir_dir exists: *)
  let tmpdir_loc =
    match !tmpdir_dir with
	None ->
	  let tmp_prefix = "/tmp/godi" ^ string_of_int (Unix.getpid()) ^ "-" in
	  let d = catch_unix_error (create_tdir tmp_prefix) 0 in
	  tmpdir_dir := Some d;
	  d
      | Some d -> d
  in
  (* Fetch next number: *)
  let n = !tmpdir_count  in
  incr tmpdir_count;
  (* Create directory: *)
  let name = Filename.concat tmpdir_loc (string_of_int n) in
  catch_unix_error (Unix.mkdir name) 0o777;
  name
;;


let release_tmpdir dir =
  remove_dir ~recurse:true dir
;;


let release_all_tmpdirs() =
  match !tmpdir_dir with
      None -> ()
    | Some d ->
	remove_dir ~recurse:true d;
	tmpdir_dir := None
;;


let copy ?(prefer_link = false) ?(remove = false) origname copyname =
  
  let do_copy =
    catch_unix_error
      (fun () ->
	 let b = String.create 4096 in
	 let origfd = open_in_bin origname in
	 try
	   if remove && Sys.file_exists copyname then
	     Sys.remove copyname;
	   let copyfd = open_out_bin copyname in
	   ( try
	       let eof = ref false in
	       while not !eof do
		 let n = input origfd b 0 (String.length b) in
		 if n=0 then eof := true;
		 output copyfd b 0 n
	       done;
	       let p = 
		 (Unix.fstat (Unix.descr_of_in_channel origfd)).Unix.st_perm in
	       Unix.fchmod (Unix.descr_of_out_channel copyfd) (p land 0o777);
	       close_in origfd;
	       close_out copyfd;
	     with
		 error ->
		   close_out copyfd; raise error
	   )
	 with
	     error -> 
	       close_in origfd; raise error
      )
  in

  (* Try to link, if demanded and possible: *)
  if prefer_link then
    try
      Unix.link origname copyname
    with
	Unix.Unix_error((Unix.EXDEV | Unix.EPERM | Unix.EMLINK), _, _) ->
	  do_copy()
      | Unix.Unix_error(_,_,_) as e ->
	  catch_unix_error (fun () -> raise e) ()
      | Invalid_argument _ ->
	  (* [link] not available *)
	  do_copy()
  else
    do_copy()
;;


let execute ?(setx=false) cmd =
  try
    let switches =
      if setx then [ "-xc" ] else [ "-c" ] in
    Unix.execv "/bin/sh" (Array.of_list
			    (["/bin/sh"] @ switches @ [cmd]));
    (* Note: Some old shells do not accept /bin/sh -x -c <cmd>, but only
     * /bin/sh -xc <cmd>!
     *)
  with
      Unix.Unix_error(err, _, param) ->
	let s = Unix.error_message err in
	let s' = if param="" then s else s ^ ": " ^ param in
	raise(Sys_error s')
;;
